@@ -401,7 +401,7 @@ class PandasEvaluator(BaseEvaluator):
             # apply reindex only if have 1 dimension. groupby work fine with sort=False with two or more dims.
             if len(_cols) == 1:
                 _groupedDF = _groupedDF.reindex(
-                    index=_result.index.get_level_values(_cols[0]).unique().to_list())
+                    index=_filteredResult.index.get_level_values(_cols[0]).unique().to_list())
 
             if useCustomFillMeasures:
                 self._applyGroupMeasures(_groupedDF, _result.groupMeasures)
@@ -496,7 +496,7 @@ class PandasEvaluator(BaseEvaluator):
         _result = self.ensureDataFrame(nodeDic[nodeId].result)
 
         if isinstance(_result, pd.DataFrame):
-            _result.to_csv(fileName, sep=columnFormat, encoding="utf-8-sig")
+            _result.to_csv(fileName, sep=columnFormat, decimal=decimalSep, encoding="utf-8-sig")
 
             return True
 
